@@ -25,12 +25,13 @@
 #include <config.h>
 #include "eel-string.h"
 
-#include <ctype.h>
 #include <errno.h>
 #include <locale.h>
 #include <stdlib.h>
 
+#if !defined (EEL_OMIT_SELF_CHECK)
 #include "eel-lib-self-check-functions.h"
+#endif
 
 size_t
 eel_strlen (const char *string)
@@ -64,8 +65,8 @@ eel_strcasecmp (const char *string_a, const char *string_b)
 	 * didn't have code that already relies on 'NULL == ""', I
 	 * would change it right now.
 	 */
-	return g_strcasecmp (string_a == NULL ? "" : string_a,
-		             string_b == NULL ? "" : string_b);
+	return g_ascii_strcasecmp (string_a == NULL ? "" : string_a,
+				   string_b == NULL ? "" : string_b);
 }
 
 int
@@ -85,34 +86,6 @@ eel_strcmp_case_breaks_ties (const char *string_a, const char *string_b)
 	return eel_strcmp (string_a, string_b);
 }
 
-int
-eel_strcoll (const char *string_a, const char *string_b)
-{
-	const char *locale;
-	int result;
-	
-	/* FIXME bugzilla.eazel.com 5450: Maybe we need to make this
-	 * treat 'NULL < ""', or have a flavor that does that. If we
-	 * didn't have code that already relies on 'NULL == ""', I
-	 * would change it right now.
-	 */
-
-	locale = setlocale (LC_COLLATE, NULL);
-	
-	if (locale == NULL || strcmp (locale, "C") == 0 || strcmp (locale, "POSIX") == 0) {
-		/* If locale is NULL or default "C" or "POSIX" use eel sorting */
-		return eel_strcmp_case_breaks_ties (string_a, string_b);
-	} else {
-		/* Use locale-specific collated sorting */
-		result = strcoll (string_a == NULL ? "" : string_a,
-				  string_b == NULL ? "" : string_b);
-		if (result != 0) {
-			return result;
-		}
-		return eel_strcmp (string_a, string_b);
-	}
-}
-
 gboolean
 eel_str_is_empty (const char *string_or_null)
 {
@@ -141,66 +114,16 @@ eel_istr_is_equal (const char *string_a, const char *string_b)
 	return eel_strcasecmp (string_a, string_b) == 0;
 }
 
-int
-eel_strcmp_compare_func (gconstpointer string_a, gconstpointer string_b)
-{
-	/* FIXME bugzilla.eazel.com 5450: Maybe we need to make this
-	 * treat 'NULL < ""', or have a flavor that does that. If we
-	 * didn't have code that already relies on 'NULL == ""', I
-	 * would change it right now.
-	 */
-	return eel_strcmp ((const char *) string_a,
-				(const char *) string_b);
-}
-
-int
-eel_strcoll_compare_func (gconstpointer string_a, gconstpointer string_b)
-{
-	/* FIXME bugzilla.eazel.com 5450: Maybe we need to make this
-	 * treat 'NULL < ""', or have a flavor that does that. If we
-	 * didn't have code that already relies on 'NULL == ""', I
-	 * would change it right now.
-	 */
-	return eel_strcoll ((const char *) string_a,
-				 (const char *) string_b);
-}
-
-int
-eel_strcasecmp_compare_func (gconstpointer string_a, gconstpointer string_b)
-{
-	/* FIXME bugzilla.eazel.com 5450: Maybe we need to make this
-	 * treat 'NULL < ""', or have a flavor that does that. If we
-	 * didn't have code that already relies on 'NULL == ""', I
-	 * would change it right now.
-	 */
-	return eel_strcasecmp ((const char *) string_a,
-				    (const char *) string_b);
-}
-
 gboolean
 eel_str_has_prefix (const char *haystack, const char *needle)
 {
-	const char *h, *n;
-
-	/* Eat one character at a time. */
-	h = haystack == NULL ? "" : haystack;
-	n = needle == NULL ? "" : needle;
-	do {
-		if (*n == '\0') {
-			return TRUE;
-		}
-		if (*h == '\0') {
-			return FALSE;
-		}
-	} while (*h++ == *n++);
-	return FALSE;
+	return g_str_has_prefix (haystack == NULL ? "" : haystack,
+				 needle == NULL ? "" : needle);
 }
 
 gboolean
 eel_str_has_suffix (const char *haystack, const char *needle)
 {
-	const char *h, *n;
-
 	if (needle == NULL) {
 		return TRUE;
 	}
@@ -208,18 +131,7 @@ eel_str_has_suffix (const char *haystack, const char *needle)
 		return needle[0] == '\0';
 	}
 		
-	/* Eat one character at a time. */
-	h = haystack + strlen(haystack);
-	n = needle + strlen(needle);
-	do {
-		if (n == needle) {
-			return TRUE;
-		}
-		if (h == haystack) {
-			return FALSE;
-		}
-	} while (*--h == *--n);
-	return FALSE;
+	return g_str_has_suffix (haystack, needle);
 }
 
 gboolean
@@ -240,8 +152,8 @@ eel_istr_has_prefix (const char *haystack, const char *needle)
 		}
 		hc = *h++;
 		nc = *n++;
-		hc = tolower ((guchar) hc);
-		nc = tolower ((guchar) nc);
+		hc = g_ascii_tolower (hc);
+		nc = g_ascii_tolower (nc);
 	} while (hc == nc);
 	return FALSE;
 }
@@ -271,8 +183,8 @@ eel_istr_has_suffix (const char *haystack, const char *needle)
 		}
 		hc = *--h;
 		nc = *--n;
-		hc = tolower ((guchar) hc);
-		nc = tolower ((guchar) nc);
+		hc = g_ascii_tolower (hc);
+		nc = g_ascii_tolower (nc);
 	} while (hc == nc);
 	return FALSE;
 }
@@ -314,44 +226,6 @@ eel_str_get_prefix (const char *source,
 	return g_strndup (source, prefix_start - source);
 }
 
-
-/**
- * eel_str_get_after_prefix:
- * Get a new string containing the part of the string
- * after the prefix
- * @source: The string whose prefix should be extracted.
- * @delimiter: The string that marks the end of the prefix.
- * 
- * Return value: A newly-allocated string that that matches the end
- * of @source, starting right after the first occurr
- * @delimiter. If @source is NULL, returns NULL. If 
- * @delimiter is NULL, returns a copy of @source.
- * If @delimiter does not occur in @source, returns
- * NULL
- **/
-char *
-eel_str_get_after_prefix (const char *source,
-			       const char *delimiter)
-{
-	char *prefix_start;
-	
-	if (source == NULL) {
-		return NULL;
-	}
-	
-	if (delimiter == NULL) {
-		return g_strdup (source);
-	}
-	
-	prefix_start = strstr (source, delimiter);
-	
-	if (prefix_start == NULL) {
-		return NULL;
-	}
-	
-	return g_strdup (prefix_start);
-}
-
 gboolean
 eel_str_to_int (const char *string, int *integer)
 {
@@ -377,7 +251,7 @@ eel_str_to_int (const char *string, int *integer)
 
 	/* Check that all the trailing characters are spaces. */
 	while (*parse_end != '\0') {
-		if (!isspace (*parse_end++)) {
+		if (!g_ascii_isspace (*parse_end++)) {
 			return FALSE;
 		}
 	}
@@ -387,96 +261,6 @@ eel_str_to_int (const char *string, int *integer)
 	return TRUE;
 }
 
-/**
- * eel_str_strip_chr:
- * Remove all occurrences of a character from a string.
- * 
- * @source: The string to be stripped.
- * @remove_this: The char to remove from @source
- * 
- * Return value: A copy of @source, after removing all occurrences
- * of @remove_this.
- */
-char *
-eel_str_strip_chr (const char *source, char remove_this)
-{
-	char *result, *out;
-	const char *in;
-	
-        if (source == NULL) {
-		return NULL;
-	}
-	
-	result = g_new (char, strlen (source) + 1);
-	in = source;
-	out = result;
-	do {
-		if (*in != remove_this) {
-			*out++ = *in;
-		}
-	} while (*in++ != '\0');
-
-        return result;
-}
-
-/**
- * eel_str_strip_trailing_chr:
- * Remove trailing occurrences of a character from a string.
- * 
- * @source: The string to be stripped.
- * @remove_this: The char to remove from @source
- * 
- * Return value: @source, after removing trailing occurrences
- * of @remove_this.
- */
-char *
-eel_str_strip_trailing_chr (const char *source, char remove_this)
-{
-	const char *end;
-	
-        if (source == NULL) {
-		return NULL;
-	}
-
-	for (end = source + strlen (source); end != source; end--) {
-		if (end[-1] != remove_this) {
-			break;
-		}
-	}
-	
-        return g_strndup (source, end - source);
-}
-
-char *   
-eel_str_strip_trailing_str (const char *source, const char *remove_this)
-{
-	const char *end;
-	if (source == NULL) {
-		return NULL;
-	}
-	if (remove_this == NULL) {
-		return g_strdup (source);
-	}
-	end = source + strlen (source);
-	if (strcmp (end - strlen (remove_this), remove_this) != 0) {
-		return g_strdup (source);
-	}
-	else {
-		return g_strndup (source, strlen (source) - strlen(remove_this));
-	}
-	
-}
-
-gboolean
-eel_eat_str_to_int (char *source, int *integer)
-{
-	gboolean result;
-
-	result = eel_str_to_int (source, integer);
-	g_free (source);
-	return result;
-}
-
 char *
 eel_str_double_underscores (const char *string)
 {
@@ -522,7 +306,7 @@ eel_str_capitalize (const char *string)
 
 	capitalized = g_strdup (string);
 
-	capitalized[0] = toupper ((guchar) capitalized[0]);
+	capitalized[0] = g_ascii_toupper (capitalized[0]);
 
 	return capitalized;
 }
@@ -537,7 +321,7 @@ eel_str_capitalize (const char *string)
  */
 char *
 eel_str_middle_truncate (const char *string,
-			      guint truncate_length)
+			 guint truncate_length)
 {
 	char *truncated;
 	guint length;
@@ -560,7 +344,7 @@ eel_str_middle_truncate (const char *string,
 		return g_strdup (string);
 	}
 
-	length = strlen (string);
+	length = g_utf8_strlen (string, -1);
 
 	/* Make sure the string is not already small enough. */
 	if (length <= truncate_length) {
@@ -569,43 +353,17 @@ eel_str_middle_truncate (const char *string,
 
 	/* Find the 'middle' where the truncation will occur. */
 	num_left_chars = (truncate_length - delimter_length) / 2;
-	num_right_chars = truncate_length - num_left_chars - delimter_length + 1;
+	num_right_chars = truncate_length - num_left_chars - delimter_length;
 
-	truncated = g_new (char, truncate_length + 1);
+	truncated = g_new (char, strlen (string) + 1);
 
-	strncpy (truncated, string, num_left_chars);
-	strncpy (truncated + num_left_chars, delimter, delimter_length);
-	strncpy (truncated + num_left_chars + delimter_length, string + length - num_right_chars + 1, num_right_chars);
+	g_utf8_strncpy (truncated, string, num_left_chars);
+	strcat (truncated, delimter);
+	strcat (truncated, g_utf8_offset_to_pointer  (string, length - num_right_chars));
 	
 	return truncated;
 }
 
-/**
- * eel_str_count_characters:
- * Count the number of 'c' characters that occur in 'string.
- * 
- * @string: The string to be scanned.
- * @c: The char to count.
- * 
- * Return value: @count, the 'c' occurance count.
- */
-guint
-eel_str_count_characters (const char	*string,
-			       char		c)
-{
-	guint count = 0;
-
-	while (string && *string != '\0') {
-		if (*string == c) {
-			count++;
-		}
-
-		string++;
-	}
-
-	return count;
-}
-
 char *
 eel_str_strip_substring_and_after (const char *string,
 					const char *substring)
@@ -678,89 +436,476 @@ eel_str_replace_substring (const char *string,
 	return result;
 }
 
+/**************** Custom printf ***********/
 
-/* Removes strings enclosed by the '[' and ']' characters.  Strings
-   that have unbalanced open and closed brackets will return the
-   string itself. */
-char *
-eel_str_remove_bracketed_text (const char *text)
+typedef struct {
+	const char *start;
+	const char *end;
+	GString *format;
+	int arg_pos;
+	int width_pos;
+	int width_format_index;
+	int precision_pos;  
+	int precision_format_index;
+} ConversionInfo;
+
+enum {
+	ARG_TYPE_INVALID,
+	ARG_TYPE_INT,
+	ARG_TYPE_LONG,
+	ARG_TYPE_LONG_LONG,
+	ARG_TYPE_SIZE,
+	ARG_TYPE_LONG_DOUBLE,
+	ARG_TYPE_DOUBLE,
+	ARG_TYPE_POINTER
+};
+
+typedef int ArgType; /* An int, because custom are < 0 */
+
+
+static const char *
+get_position (const char *format, int *i)
 {
-	char *unbracketed_text;
-	char *unbracketed_segment, *new_unbracketed_text;
-	const char *current_text_location;
-	const char *next_open_bracket, *next_close_bracket;
-	int bracket_depth;
+	const char *p;
 	
+	p = format;
+	
+	if (g_ascii_isdigit (*p)) {
+		p++;
+		
+		while (g_ascii_isdigit (*p)) {
+			p++;
+		}
+		
+		if (*p == '$') {
+			if (i != NULL) {
+				*i = atoi (format) - 1;
+			}
+			return p + 1;
+		}
+	}
+	
+	return format;
+}
 
-	g_return_val_if_fail (text != NULL, NULL);
+static gboolean
+is_flag (char c)
+{
+	return strchr ("#0- +'I", c) != NULL;
+}
 
-	current_text_location = text;
-	bracket_depth = 0;
-	unbracketed_text = g_strdup ("");
-	while (TRUE) {
-		next_open_bracket = strchr (current_text_location, '[');
-		next_close_bracket = strchr (current_text_location, ']');
-		/* No more brackets */
-		if (next_open_bracket == NULL &&
-		    next_close_bracket == NULL) {
-			if (bracket_depth == 0) {
-				new_unbracketed_text = g_strconcat (unbracketed_text, 
-								    current_text_location, NULL);
-				g_free (unbracketed_text);
-				return new_unbracketed_text;
-			}
-			else {
-				g_free (unbracketed_text);
-				return g_strdup (text);
-			}
-		}
-		/* Close bracket but no open bracket */
-		else if (next_open_bracket == NULL) {
-			if (bracket_depth == 0) {
-				g_free (unbracketed_text);
-				return g_strdup (text);
-			}
-			else {
-				current_text_location = next_close_bracket + 1;
-				bracket_depth--;
-			}
-		}
-		/* Open bracket but no close bracket */
-		else if (next_close_bracket == NULL) {
-			g_free (unbracketed_text);
-			return g_strdup (text);
-		}
-		/* Deal with the open bracket, that's first */
-		else if (next_open_bracket < next_close_bracket) {
-			if (bracket_depth == 0) {
-				/* We're out of brackets. Copy until the next bracket */
-				unbracketed_segment = g_strndup (current_text_location,
-								 next_open_bracket - current_text_location);
-				new_unbracketed_text = g_strconcat (unbracketed_text, unbracketed_segment, NULL);
-				g_free (unbracketed_text);
-				g_free (unbracketed_segment);
-				unbracketed_text = new_unbracketed_text;
-			}
-			current_text_location = next_open_bracket + 1;
-			bracket_depth++;
-		}
-		/* Deal with the close bracket, that's first */
-		else {
-			if (bracket_depth > 0) {
-				bracket_depth--;
-				current_text_location = next_close_bracket + 1;
-			}
-			else {
-				g_free (unbracketed_text);
-				return g_strdup (text);
+static gboolean
+is_length_modifier (char c)
+{
+	return strchr ("hlLjzt", c) != NULL;
+}
+
+
+static ArgType
+get_arg_type_from_format (EelPrintfHandler *custom_handlers,
+			  const char *format,
+			  int len)
+{
+	int i;
+	char c;
+	
+	c = format[len-1];
+	
+	if (custom_handlers != NULL) {
+		for (i = 0; custom_handlers[i].character != 0; i++) {
+			if (custom_handlers[i].character == c) {
+				return -(i + 1);
 			}
 		}
 	}
 	
-	     
-
+	switch (c) {
+	case 'd':
+	case 'i':
+	case 'o':
+	case 'u':
+	case 'x':
+	case 'X':
+		if (g_str_has_prefix (format, "ll")) {
+			return ARG_TYPE_LONG_LONG;
+		}
+		if (g_str_has_prefix (format, "l")) {
+			return ARG_TYPE_LONG;
+		}
+		if (g_str_has_prefix (format, "l")) {
+			return ARG_TYPE_LONG;
+		}
+		if (g_str_has_prefix (format, "z")) {
+			return ARG_TYPE_SIZE;
+		}
+		return ARG_TYPE_INT;
+	case 'e':
+	case 'E':
+	case 'f':
+	case 'F':
+	case 'g':
+	case 'G':
+	case 'a':
+	case 'A':
+		if (g_str_has_prefix (format, "L")) {
+			return ARG_TYPE_LONG_DOUBLE;
+		}
+		return ARG_TYPE_DOUBLE;
+	case 'c':
+		return ARG_TYPE_INT;
+	case 's':
+	case 'p':
+	case 'n':
+		return ARG_TYPE_POINTER;
+	}
+	return ARG_TYPE_INVALID;
 }
 
+static void
+skip_argv (va_list *va,
+	   ArgType type,
+	   EelPrintfHandler *custom_handlers)
+{
+	if (type < 0) {
+		custom_handlers[-type - 1].skip (va);
+		return;
+	}
+	
+	switch (type) {
+	default:
+	case ARG_TYPE_INVALID:
+		return;
+		
+	case ARG_TYPE_INT:
+		va_arg (*va, int);
+		break;
+	case ARG_TYPE_LONG:
+		va_arg (*va, long int);
+		break;
+	case ARG_TYPE_LONG_LONG:
+		va_arg (*va, long long int);
+		break;
+	case ARG_TYPE_SIZE:
+		va_arg (*va, gsize);
+		break;
+	case ARG_TYPE_LONG_DOUBLE:
+		va_arg (*va, long double);
+		break;
+	case ARG_TYPE_DOUBLE:
+		va_arg (*va, double);
+		break;
+	case ARG_TYPE_POINTER:
+		va_arg (*va, void *);
+		break;
+	}
+}
+
+static void
+skip_to_arg (va_list *va,
+	     ArgType *types,
+	     EelPrintfHandler *custom_handlers,
+	     int n)
+{
+	int i;
+	for (i = 0; i < n; i++) {
+		skip_argv (va, types[i], custom_handlers);
+	}
+}
+
+char *
+eel_strdup_vprintf_with_custom (EelPrintfHandler *custom,
+				const char *format,
+				va_list va_orig)
+{
+	va_list va;
+	const char *p;
+	int num_args, i, j;
+	ArgType *args;
+	ArgType type;
+	ConversionInfo *conversions;
+	GString *f, *str;
+	const char *flags, *width, *prec, *mod, *pos;
+	char *s;
+	
+	num_args = 0;
+	for (p = format; *p != 0; p++) {
+		if (*p == '%') {
+			p++;
+			if (*p != '%') {
+				num_args++;
+			}
+		}
+	}
+	
+	args = g_new0 (ArgType, num_args * 3 + 1);
+	conversions = g_new0 (ConversionInfo, num_args);
+	
+	/* i indexes conversions, j indexes args */
+	i = 0; j = 0;
+	p = format;
+	while (*p != 0) {
+		if (*p != '%') {
+			p++;
+			continue;
+		}
+		p++;
+		if (*p == '%') {
+			p++;
+			continue;
+		}
+		
+		/* We got a real conversion: */
+		f = g_string_new ("%");
+		conversions[i].start = p - 1;
+		
+		/* First comes the positional arg */
+		
+		pos = p;
+		p = get_position (p, NULL);
+		
+		/* Then flags */
+		flags = p;
+		while (is_flag (*p)) {
+			p++;
+		}
+		g_string_append_len (f, flags, p - flags);
+		
+		/* Field width */
+		
+		if (*p == '*') {
+			p++;
+			p = get_position (p, &j);
+			args[j] = ARG_TYPE_INT;
+			conversions[i].width_pos = j++;
+			conversions[i].width_format_index = f->len;
+		} else {
+			conversions[i].width_pos = -1;
+			conversions[i].width_format_index = -1;
+			width = p;
+			while (g_ascii_isdigit (*p)) {
+				p++;
+			}      
+			g_string_append_len (f, width, p - width);
+		}
+		
+		/* Precision */
+		conversions[i].precision_pos = -1;
+		conversions[i].precision_format_index = -1;
+		if (*p == '.') {
+			g_string_append_c (f, '.');
+			p++;
+			
+			if (*p == '*') {
+				p++;
+				p = get_position (p, &j);
+				args[j] = ARG_TYPE_INT;
+				conversions[i].precision_pos = j++;
+				conversions[i].precision_format_index = f->len;
+			} else {
+				prec = p;
+				while (g_ascii_isdigit (*p) || *p == '-') {
+					p++;
+				}      
+				g_string_append_len (f, prec, p - prec);
+			}
+		}
+		
+		/* length modifier */
+		
+		mod = p;
+		
+		while (is_length_modifier (*p)) {
+			p++;
+		}
+		
+		/* conversion specifier */
+		if (*p != 0)
+			p++;
+		
+		g_string_append_len (f, mod, p - mod);
+		
+		get_position (pos, &j);
+		args[j] = get_arg_type_from_format (custom, mod, p - mod);
+		conversions[i].arg_pos = j++;
+		conversions[i].format = f;
+		conversions[i].end = p;
+		
+		i++;
+	}
+	
+	g_assert (i == num_args);
+	
+	str = g_string_new ("");
+	
+	p = format;
+	for (i = 0; i < num_args; i++) {
+		g_string_append_len (str, p, conversions[i].start - p);
+		p = conversions[i].end;
+		
+		if (conversions[i].precision_pos != -1) {
+			char *val;
+			
+			va_copy(va, va_orig);
+			skip_to_arg (&va, args, custom, conversions[i].precision_pos);
+			val = g_strdup_vprintf ("%d", va);
+			va_end (va);
+			
+			g_string_insert (conversions[i].format,
+					 conversions[i].precision_format_index,
+					 val);
+			
+			g_free (val);
+		}
+		
+		if (conversions[i].width_pos != -1) {
+			char *val;
+			
+			va_copy(va, va_orig);
+			skip_to_arg (&va, args, custom, conversions[i].width_pos);
+			val = g_strdup_vprintf ("%d", va);
+			va_end (va);
+			
+			g_string_insert (conversions[i].format,
+					 conversions[i].width_format_index,
+					 val);
+			
+			g_free (val);
+		}
+		
+		va_copy(va, va_orig);
+		skip_to_arg (&va, args, custom, conversions[i].arg_pos);
+		type = args[conversions[i].arg_pos];
+		if (type < 0) {
+			s = custom[-type - 1].to_string (conversions[i].format->str, va);
+			g_string_append (str, s);
+			g_free (s);
+		} else{
+			g_string_append_vprintf (str, conversions[i].format->str, va);
+		}
+		va_end (va);
+		
+		g_string_free (conversions[i].format, TRUE);
+	}
+	g_string_append (str, p);
+	
+	g_free (args);
+	g_free (conversions);
+	
+	return g_string_free (str, FALSE);
+}
+
+char *
+eel_strdup_printf_with_custom (EelPrintfHandler *handlers,
+			       const char *format,
+			       ...)
+{
+	va_list va;
+	char *res;
+	
+	va_start (va, format);
+	res = eel_strdup_vprintf_with_custom (handlers, format, va);
+	va_end (va);
+	
+	return res;
+}
+
+/*********** refcounted strings ****************/
+
+G_LOCK_DEFINE_STATIC (unique_ref_strs);
+static GHashTable *unique_ref_strs = NULL;
+
+static eel_ref_str
+eel_ref_str_new_internal (const char *string, int start_count)
+{
+	char *res;
+	volatile gint *count;
+	gsize len;
+
+	len = strlen (string);
+	res = g_malloc (sizeof (gint) + len + 1);
+	count = (volatile gint *)res;
+	*count = start_count;
+	res += sizeof(gint);
+	memcpy (res, string, len + 1);
+	return res;
+}
+
+eel_ref_str
+eel_ref_str_new (const char *string)
+{
+	if (string == NULL) {
+		return NULL;
+	}
+	
+	return eel_ref_str_new_internal (string, 1);
+}
+
+eel_ref_str
+eel_ref_str_get_unique (const char *string)
+{
+	eel_ref_str res;
+
+	if (string == NULL) {
+		return NULL;
+	}
+	
+	G_LOCK (unique_ref_strs);
+	if (unique_ref_strs == NULL) {
+		unique_ref_strs = g_hash_table_new (g_str_hash, g_str_equal);
+	}
+
+	res = g_hash_table_lookup (unique_ref_strs, string);
+	if (res != NULL) {
+		eel_ref_str_ref (res);
+	} else {
+		res = eel_ref_str_new_internal (string, 0x80000001);
+		g_hash_table_insert (unique_ref_strs, res, res);
+	}
+	
+	G_UNLOCK (unique_ref_strs);
+
+	return res;
+}
+
+eel_ref_str
+eel_ref_str_ref (eel_ref_str str)
+{
+	volatile gint *count;
+
+	count = (volatile gint *)((char *)str - sizeof (gint));
+	g_atomic_int_add (count, 1);
+
+	return str;
+}
+
+void
+eel_ref_str_unref (eel_ref_str str)
+{
+	volatile gint *count;
+	gint old_ref;
+
+	if (str == NULL)
+		return;
+	
+	count = (volatile gint *)((char *)str - sizeof (gint));
+
+	old_ref = g_atomic_int_get (count);
+	if (old_ref == 1) {
+		g_free ((char *)count);
+	} else if (old_ref == 0x80000001) {
+		G_LOCK (unique_ref_strs);
+		/* Need to recheck after taking lock to avoid races with _get_unique() */
+		if (g_atomic_int_exchange_and_add (count, -1) == 0x80000001) {
+			g_hash_table_remove (unique_ref_strs, (char *)str);
+			g_free ((char *)count);
+		} 
+		G_UNLOCK (unique_ref_strs);
+	} else {
+		g_atomic_int_exchange_and_add (count, -1);
+	}
+}
+
+
 #if !defined (EEL_OMIT_SELF_CHECK)
 
 static int
@@ -773,14 +918,74 @@ call_str_to_int (const char *string)
 	return integer;
 }
 
-static int
-call_eat_str_to_int (char *string)
+static void
+verify_printf (const char *format, ...)
 {
-	int integer;
+	va_list va;
+	char *orig, *new;
+	
+	va_start (va, format);
+	orig = g_strdup_vprintf (format, va);
+	va_end (va);
+	
+	va_start (va, format);
+	new = eel_strdup_vprintf_with_custom (NULL, format, va);
+	va_end (va);
+	
+	EEL_CHECK_STRING_RESULT (new, orig);
+	
+	g_free (orig);
+}
 
-	integer = 9999;
-	eel_eat_str_to_int (string, &integer);
-	return integer;
+static char *
+custom1_to_string (char *format, va_list va)
+{
+	int i;
+	
+	i = va_arg (va, int);
+	
+	return g_strdup_printf ("c1-%d-", i);
+}
+
+static void
+custom1_skip (va_list *va)
+{
+	va_arg (*va, int);
+}
+
+static char *
+custom2_to_string (char *format, va_list va)
+{
+	char *s;
+	
+	s = va_arg (va, char *);
+	
+	return g_strdup_printf ("c2-%s-", s);
+}
+
+static void
+custom2_skip (va_list *va)
+{
+	va_arg (*va, char *);
+}
+
+static EelPrintfHandler handlers[] = {
+	{ 'N', custom1_to_string, custom1_skip },
+	{ 'Y', custom2_to_string, custom2_skip },
+	{ 0 }
+};
+
+static void
+verify_custom (const char *orig, const char *format, ...)
+{
+	char *new;
+	va_list va;
+	
+	va_start (va, format);
+	new = eel_strdup_vprintf_with_custom (handlers, format, va);
+	va_end (va);
+	
+	EEL_CHECK_STRING_RESULT (new, orig);
 }
 
 void
@@ -852,43 +1057,6 @@ eel_self_check_string (void)
 	EEL_CHECK_STRING_RESULT (eel_str_get_prefix ("foo:bar", ":"), "foo");
 	EEL_CHECK_STRING_RESULT (eel_str_get_prefix ("footle:bar", "tle:"), "foo");	
 
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix (NULL, NULL), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix (NULL, "foo"), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("foo", NULL), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("", ""), "");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("", "foo"), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("foo", ""), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("foo", "foo"), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("foo:", ":"), ":");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("foo:bar", ":"), ":bar");
-	EEL_CHECK_STRING_RESULT (eel_str_get_after_prefix ("footle:bar", "tle:"), "tle:bar");	
-
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr (NULL, '_'), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("", '_'), "");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("foo", '_'), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("_foo", '_'), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("foo_", '_'), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("_foo__", '_'), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_chr ("_f_o__o_", '_'), "foo");
-        
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr (NULL, '_'), NULL);	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("", '_'), "");	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("foo", '_'), "foo");	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("_foo", '_'), "_foo");	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("foo_", '_'), "foo");	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("_foo__", '_'), "_foo");	
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_chr ("_f_o__o_", '_'), "_f_o__o");	
-
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str (NULL, NULL), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str (NULL, "bar"), NULL);
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("bar", NULL), "bar");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("", ""), "");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("", "bar"), "");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("bar", ""), "bar");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("foo", "bar"), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("foo bar", "bar"), "foo ");
-	EEL_CHECK_STRING_RESULT (eel_str_strip_trailing_str ("bar", "bar"), "");
-	
 	EEL_CHECK_STRING_RESULT (eel_str_double_underscores (NULL), NULL);
 	EEL_CHECK_STRING_RESULT (eel_str_double_underscores (""), "");
 	EEL_CHECK_STRING_RESULT (eel_str_double_underscores ("_"), "__");
@@ -937,9 +1105,7 @@ eel_self_check_string (void)
 
 	#define TEST_INTEGER_CONVERSION_FUNCTIONS(string, boolean_result, integer_result) \
 		EEL_CHECK_BOOLEAN_RESULT (eel_str_to_int (string, &integer), boolean_result); \
-		EEL_CHECK_INTEGER_RESULT (call_str_to_int (string), integer_result); \
-		EEL_CHECK_BOOLEAN_RESULT (eel_eat_str_to_int (g_strdup (string), &integer), boolean_result); \
-		EEL_CHECK_INTEGER_RESULT (call_eat_str_to_int (g_strdup (string)), integer_result);
+		EEL_CHECK_INTEGER_RESULT (call_str_to_int (string), integer_result);
 
 	TEST_INTEGER_CONVERSION_FUNCTIONS (NULL, FALSE, 9999)
 	TEST_INTEGER_CONVERSION_FUNCTIONS ("", FALSE, 9999)
@@ -980,15 +1146,6 @@ eel_self_check_string (void)
 	EEL_CHECK_BOOLEAN_RESULT (eel_istr_is_equal ("Foo", "foo"), TRUE);
 	EEL_CHECK_BOOLEAN_RESULT (eel_istr_is_equal ("foo", "Foo"), TRUE);
 
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters (NULL, 'x'), 0);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("", 'x'), 0);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters (NULL, '\0'), 0);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("", '\0'), 0);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("foo", 'x'), 0);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("foo", 'f'), 1);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("foo", 'o'), 2);
-	EEL_CHECK_INTEGER_RESULT (eel_str_count_characters ("xxxx", 'x'), 4);
-
 	EEL_CHECK_STRING_RESULT (eel_str_strip_substring_and_after (NULL, "bar"), NULL);
 	EEL_CHECK_STRING_RESULT (eel_str_strip_substring_and_after ("", "bar"), "");
 	EEL_CHECK_STRING_RESULT (eel_str_strip_substring_and_after ("foo", "bar"), "foo");
@@ -1007,19 +1164,14 @@ eel_self_check_string (void)
 	EEL_CHECK_STRING_RESULT (eel_str_replace_substring ("foofoofoo", "foo", "f"), "fff");
 	EEL_CHECK_STRING_RESULT (eel_str_replace_substring ("foofoofoo", "f", ""), "oooooo");
 
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text (""), "");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("[]"), "");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("["), "[");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("]"), "]");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("[[]"), "[[]");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo"), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo [bar]"), "foo ");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ bar]"), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ bar] baz"), "foo baz");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ [b]ar] baz"), "foo baz");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ bar] baz[ bat]"), "foo baz");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ bar[ baz] bat]"), "foo");
-	EEL_CHECK_STRING_RESULT (eel_str_remove_bracketed_text ("foo[ bar] baz] bat]"), "foo[ bar] baz] bat]");
+	verify_printf ("%.*s", 2, "foo");
+	verify_printf ("%*.*s", 2, 4, "foo");
+	verify_printf ("before %5$*1$.*2$s between %6$*3$.*4$d after",
+		       4, 5, 6, 7, "foo", G_PI);
+	verify_custom ("c1-42- c2-foo-","%N %Y", 42 ,"foo");
+	verify_custom ("c1-42- bar c2-foo-","%N %s %Y", 42, "bar" ,"foo");
+	verify_custom ("c1-42- bar c2-foo-","%3$N %2$s %1$Y","foo", "bar", 42);
+
 }
 
 #endif /* !EEL_OMIT_SELF_CHECK */
